"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StepFunctionsOfflinePlugin = void 0;
const child_process_1 = require("child_process");
const dayjs = require("dayjs");
const fs = require("fs");
const lodash_1 = require("lodash");
const os = require("os");
const path = require("path");
const uuid_1 = require("uuid");
const enum_1 = require("./enum");
const types_1 = require("./types");
const worker_1 = require("./worker");
const delay = (time) => new Promise((resolve) => setTimeout(resolve, time * 1000));
const isString = (item) => typeof item == 'string';
const fileReferenceRegex = /\$\{file\((.+)\)\}$/;
function serialize(any) {
    return JSON.parse(JSON.stringify(any));
}
class StepFunctionsOfflinePlugin {
    constructor(serverless, options, cli) {
        var _a, _b;
        this.environmentVariables = {};
        this.mapResults = [];
        this.states = {};
        this.eventParallelResult = [];
        this.shouldTerminate = true;
        this.environment = '';
        this.location = process.cwd();
        this.serverless = serverless;
        this.cli = cli;
        this.options = options;
        this.stateMachine = this.options.stateMachine;
        this.detailedLog = (_a = (this.options.detailedLog || this.options.l)) !== null && _a !== void 0 ? _a : false;
        this.event = this.options.event || this.options.e;
        this.eventFile = this.options.eventFile || this.options.ef;
        this.functions = this.serverless.service.functions;
        this.variables = (_b = this.serverless.service.custom) === null || _b === void 0 ? void 0 : _b.stepFunctionsOffline;
        this.cliLog = this.cli.log.info.bind(this.cli);
        this.contexts = {};
        this.commands = {
            'step-functions-offline': {
                usage: 'Will run your step function locally',
                lifecycleEvents: [
                    'checkVariableInYML',
                    'start',
                    'isInstalledPluginSLSStepFunctions',
                    'findFunctionsPathAndHandler',
                    'findState',
                    'loadEventFile',
                    'loadEnvVariables',
                    'buildStepWorkFlow',
                    'exit',
                ],
                options: {
                    stateMachine: {
                        usage: 'The stage used to execute.',
                        required: true,
                    },
                    eventFile: {
                        usage: 'File where is values for execution in JSON format',
                        shortcut: 'ef',
                    },
                    event: {
                        usage: 'Values in JSON format',
                        shortcut: 'e',
                    },
                    detailedLog: {
                        usage: 'Option which enables detailed logs',
                        shortcut: 'l',
                    },
                },
            },
        };
        this.hooks = {
            'step-functions-offline:start': this.start.bind(this),
            'step-functions-offline:isInstalledPluginSLSStepFunctions': this.isInstalledPluginSLSStepFunctions.bind(this),
            'step-functions-offline:findState': this.findState.bind(this),
            'step-functions-offline:loadEventFile': this.loadEventFile.bind(this),
            'step-functions-offline:loadEnvVariables': this.loadEnvVariables.bind(this),
            'step-functions-offline:buildStepWorkFlow': this.buildStepWorkFlow.bind(this),
            'step-functions-offline:exit': this.exit.bind(this),
        };
    }
    exit() {
        if (this.shouldTerminate)
            process.exit(0);
    }
    // Entry point for the plugin (sls step offline)
    start() {
        this.cliLog('Preparing....');
        this._getLocation();
        this._checkVersion();
        this._checkVariableInYML();
    }
    _getLocation() {
        if (this.options.location) {
            this.location = path.join(process.cwd(), this.options.location);
        }
        if (this.variables && this.variables.location) {
            this.location = path.join(process.cwd(), this.variables.location);
        }
    }
    _checkVersion() {
        const version = this.serverless.version;
        if (!version.startsWith('3.')) {
            throw new Error(`Serverless Step Functions Offline requires Serverless v3.x.x but found ${version}`);
        }
    }
    _checkVariableInYML() {
        if (!(0, lodash_1.has)(this.serverless.service, 'custom.stepFunctionsOffline')) {
            throw new Error('Please add ENV_VARIABLES to section "custom"');
        }
        return;
    }
    isInstalledPluginSLSStepFunctions() {
        const plugins = this.serverless.service.plugins;
        if (plugins.indexOf('serverless-step-functions') < 0) {
            const error = 'Error: Please install plugin "serverless-step-functions". Package does not work without it';
            throw new Error(error);
        }
    }
    loadEventFile() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.eventFile) {
                this.loadedEventFile = this.event;
                return Promise.resolve();
            }
            try {
                this.loadedEventFile = path.isAbsolute(this.eventFile)
                    ? yield Promise.resolve().then(() => require(this.eventFile))
                    : yield Promise.resolve().then(() => require(path.join(process.cwd(), this.eventFile)));
            }
            catch (err) {
                throw err;
            }
        });
    }
    loadEnvVariables() {
        // this.environment = this.serverless.service.provider.environment;
        process.env.STEP_IS_OFFLINE = 'true';
        process.env = (0, lodash_1.extend)(process.env, this.environment);
        this.environmentVariables = Object.assign({}, process.env); //store global env variables;
        return;
    }
    findState() {
        this.cliLog(`Trying to find state "${this.stateMachine}" in serverless manifest`);
        return this.parseConfig()
            .then(() => {
            if (!this.stateMachine) {
                throw new Error('unable to get state definition');
            }
            this.stateDefinition = this.getStateMachine(this.stateMachine).definition;
        })
            .catch((err) => {
            throw new Error(err.message);
        });
    }
    parseYaml(filename) {
        return this.serverless.yamlParser.parse(filename);
    }
    serverlessFileExists(filename) {
        const serverlessPath = this.serverless.config.servicePath;
        if (!serverlessPath) {
            throw new Error('Could not find serverless manifest');
        }
        const fullPath = path.join(serverlessPath, filename);
        return this.serverless.utils.fileExistsSync(fullPath);
    }
    getRawConfig() {
        return __awaiter(this, void 0, void 0, function* () {
            const serverlessPath = this.serverless.config.servicePath;
            if (!serverlessPath) {
                throw new Error('Could not find serverless manifest');
            }
            const manifestFilenames = ['serverless.yaml', 'serverless.yml', 'serverless.json', 'serverless.js'];
            const manifestFilename = manifestFilenames.find((name) => this.serverlessFileExists(name));
            if (!manifestFilename) {
                throw new Error(`Could not find serverless manifest at path ${serverlessPath}. If this path is incorreect you should adjust the 'servicePath' variable`);
            }
            const manifestPath = path.join(serverlessPath, manifestFilename);
            let fromFile;
            if (/\.json|\.js$/.test(manifestPath)) {
                try {
                    fromFile = yield Promise.resolve().then(() => require(manifestPath));
                    return fromFile;
                }
                catch (err) {
                    console.error(err);
                    throw new Error(`Unable to import manifest at: ${manifestPath}`);
                }
            }
            return this.parseYaml(manifestPath);
        });
    }
    parseConfig() {
        return this.getRawConfig().then((serverlessFileParam) => {
            var _a, _b, _c, _d;
            this.serverless.service['stepFunctions'] = {};
            this.serverless.service['stepFunctions']['stateMachines'] =
                serverlessFileParam['stepFunctions'] && ((_a = serverlessFileParam['stepFunctions']) === null || _a === void 0 ? void 0 : _a.stateMachines)
                    ? (_b = serverlessFileParam['stepFunctions']) === null || _b === void 0 ? void 0 : _b.stateMachines
                    : {};
            this.serverless.service['stepFunctions']['activities'] =
                serverlessFileParam['stepFunctions'] && ((_c = serverlessFileParam['stepFunctions']) === null || _c === void 0 ? void 0 : _c.activities)
                    ? (_d = serverlessFileParam['stepFunctions']) === null || _d === void 0 ? void 0 : _d.activities
                    : [];
            if (!this.serverless.pluginManager.cliOptions['stage']) {
                this.serverless.pluginManager.cliOptions['stage'] =
                    this.options.stage || (this.serverless.service.provider && this.serverless.service.provider.stage) || 'dev';
            }
            if (!this.serverless.pluginManager.cliOptions['region']) {
                this.serverless.pluginManager.cliOptions['region'] =
                    this.options.region ||
                        (this.serverless.service.provider && this.serverless.service.provider.region) ||
                        'us-east-1';
            }
            if (this.serverless.variables.populateService) {
                this.serverless.variables.populateService();
            }
            return Promise.resolve();
        });
    }
    getStateMachine(stateMachineName) {
        var _a, _b, _c;
        if (((_a = this.serverless.service['stepFunctions']) === null || _a === void 0 ? void 0 : _a.stateMachines) &&
            stateMachineName in ((_b = this.serverless.service['stepFunctions']) === null || _b === void 0 ? void 0 : _b.stateMachines)) {
            return (_c = this.serverless.service['stepFunctions']) === null || _c === void 0 ? void 0 : _c.stateMachines[stateMachineName];
        }
        throw new Error(`stateMachine "${stateMachineName}" doesn't exist in this Service`);
    }
    _loadStates(states) {
        return __awaiter(this, void 0, void 0, function* () {
            if (isString(states)) {
                const serverlessPath = this.serverless.config.servicePath;
                if (!serverlessPath) {
                    throw new Error('Could not find serverless manifest');
                }
                const match = states.match(fileReferenceRegex);
                if (!match) {
                    throw new Error(`couldn't understand string of States: ${states}`);
                }
                const fileName = match[1];
                if (!this.serverlessFileExists(fileName)) {
                    throw new Error(`Unable to find ${fileName} in serverless path`);
                }
                return this.parseYaml(path.join(serverlessPath, fileName));
            }
            return states;
        });
    }
    buildStepWorkFlow() {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            this.cliLog('Building StepWorkFlow');
            if (!this.stateDefinition)
                throw new Error('Missing state definition');
            const event = (_a = this.loadedEventFile) !== null && _a !== void 0 ? _a : {};
            if (!((_b = this.stateDefinition) === null || _b === void 0 ? void 0 : _b.StartAt)) {
                throw new Error('Missing `startAt` in definition');
            }
            if (typeof this.stateDefinition.States === 'string') {
                this.stateDefinition.States = yield this._loadStates(this.stateDefinition.States);
            }
            this.addContextObject(this.stateDefinition.States, this.stateDefinition.StartAt, event);
            this.states = this.stateDefinition.States;
            return this.process(this.states[this.stateDefinition.StartAt], this.stateDefinition.StartAt, event);
        });
    }
    buildSubStepWorkFlow(stateDefinition, event) {
        return __awaiter(this, void 0, void 0, function* () {
            this.cliLog('Building Iterator StepWorkFlow');
            if (typeof stateDefinition.States === 'string') {
                stateDefinition.States = yield this._loadStates(stateDefinition.States);
            }
            this.addContextObject(stateDefinition.States, stateDefinition.StartAt, event);
            if (!stateDefinition.States)
                return;
            const state = stateDefinition.States[stateDefinition.StartAt];
            const result = yield this.process(state, stateDefinition.StartAt, event);
            this.contexts[stateDefinition.StartAt] = null;
            return result;
        });
    }
    process(state, stateName, event) {
        if (state && state.Type === 'Parallel') {
            this.eventForParallelExecution = event;
        }
        const data = this._findStep(state, stateName);
        // if (data instanceof Promise) return Promise.resolve();
        if (!data || data instanceof Promise) {
            if ((!state || state.Type !== 'Parallel') && !this.mapResults) {
                this.cliLog('Serverless step function offline: Finished');
            }
            return Promise.resolve();
        }
        if ((0, types_1.stateIsChoiceConditional)(data) && data.choice) {
            return this._runChoice(data, event);
        }
        else if (!(0, types_1.stateIsChoiceConditional)(data)) {
            const callback = data.f(event);
            return this._run(callback, event, stateName);
        }
    }
    _findStep(currentState, currentStateName) {
        // it means end of states
        if (!currentState) {
            this.currentState = null;
            this.currentStateName = null;
            return;
        }
        this.currentState = currentState;
        this.currentStateName = currentStateName;
        return this._states(currentState, currentStateName);
    }
    _run(func, event, name) {
        if (!func)
            return; // end of states
        this.executionLog(`~~~~~~~~~~~~~~~~~~~~~~~~~~~ ${this.currentStateName} started ~~~~~~~~~~~~~~~~~~~~~~~~~~~`);
        const contextObject = this.contexts[name];
        if (contextObject) {
            if (func instanceof Promise) {
                return func.then((mod) => __awaiter(this, void 0, void 0, function* () {
                    if (!mod)
                        return contextObject.done(null, {});
                    let res;
                    let err;
                    try {
                        const done = (e, val) => {
                            res = val;
                            err = e;
                        };
                        const functionRes = yield mod(event, contextObject, done);
                        if (functionRes)
                            res = functionRes;
                        try {
                            if (typeof res === 'string')
                                res = JSON.parse(res);
                        }
                        catch (err) { }
                    }
                    catch (error) {
                        err = error;
                    }
                    return contextObject.done(err, res || {});
                }));
            }
            return func(event, contextObject, contextObject.done);
        }
    }
    _handleMap(currentState, stateName) {
        return {
            f: (event) => {
                var _a, _b;
                const items = (0, lodash_1.get)(event, (_b = (_a = currentState.ItemsPath) === null || _a === void 0 ? void 0 : _a.replace(/^\$\./, '')) !== null && _b !== void 0 ? _b : '', []);
                const mapItems = (0, lodash_1.clone)(items);
                this.mapResults = [];
                if (mapItems.length === 0) {
                    this.cliLog(`State ${stateName} is being called with no items, skipping...`);
                    if (currentState.End)
                        return Promise.resolve();
                }
                const concurrency = currentState.MaxConcurrency || 1;
                const executeMapperPromise = (0, worker_1.withWorkers)(mapItems, (item) => __awaiter(this, void 0, void 0, function* () {
                    if (!item)
                        return;
                    const parseValue = (value) => {
                        if (value === '$$.Map.Item.Value') {
                            return item;
                        }
                        if (/^\$\./.test(value)) {
                            return (0, lodash_1.get)(event, value.replace(/^\$\./, ''));
                        }
                    };
                    const newEvent = currentState.Parameters
                        ? Object.keys(currentState.Parameters).reduce((acc, key) => {
                            if (/\.\$$/.test(key) && currentState.Parameters) {
                                acc[key.replace(/\.\$$/, '')] = parseValue(currentState.Parameters[key].toString());
                            }
                            return acc;
                        }, {})
                        : {};
                    return this.buildSubStepWorkFlow(currentState.Iterator, newEvent);
                }), concurrency, (v) => this.executionLog(v));
                return executeMapperPromise.then(() => __awaiter(this, void 0, void 0, function* () {
                    const mappedResult = yield Promise.all(this.mapResults);
                    if (currentState.ResultPath) {
                        (0, lodash_1.set)(event, currentState.ResultPath.replace(/\$\./, ''), serialize(mappedResult));
                    }
                    this.mapResults = [];
                    if (currentState.Next) {
                        this.addContextObject(this.states, currentState.Next, event);
                        yield this.process(this.states[currentState.Next], currentState.Next, event);
                    }
                    return;
                }));
            },
        };
    }
    _invokeLambdaFn(func) {
        const env = (0, lodash_1.extend)(process.env, func.environment);
        return (event) => __awaiter(this, void 0, void 0, function* () {
            return (_, _context) => {
                const processId = (0, uuid_1.v4)();
                const response = path.join(os.tmpdir(), `${processId}.json`);
                const input = path.join(os.tmpdir(), `${processId}-input.json`);
                fs.writeFileSync(input, JSON.stringify(event));
                const cmd = (0, child_process_1.spawn)('aws', [
                    'lambda',
                    'invoke',
                    response,
                    '--endpoint-url',
                    this.lambdaEndpoint,
                    '--function-name',
                    func.name,
                    '--payload',
                    `fileb://${input}`,
                ], { env });
                return new Promise((r, f) => {
                    cmd.on('close', (code) => {
                        if (code !== 0) {
                            f(new Error('Lambda invocation failed with code ' + code));
                            return;
                        }
                        try {
                            const value = fs.readFileSync(response).toString();
                            try {
                                const json = JSON.parse(value || '{}');
                                r(json);
                            }
                            catch (err) {
                                f(new Error('Fail to parse lambda response: ' + value));
                            }
                        }
                        catch (err) {
                            f(new Error('Fail to read lambda response from file: ' + response));
                        }
                        return;
                    });
                });
            };
        });
    }
    _handleTask(currentState, stateName) {
        var _a;
        // just push task to general array
        //before each task restore global default env variables
        process.env = Object.assign({}, this.environmentVariables);
        const functionName = (_a = this.variables) === null || _a === void 0 ? void 0 : _a[stateName];
        let f = functionName ? this.functions[functionName] : null;
        if (f === null && Array.isArray(this.functions)) {
            f = functionName ? this.functions.find((func) => func[functionName])[functionName] : null;
        }
        if (f === undefined || f === null) {
            this.cliLog(`Function "${stateName}" does not presented in serverless manifest`);
            throw new Error(`Function "${stateName}" does not presented in serverless manifest`);
        }
        if (!(0, types_1.definitionIsHandler)(f))
            return;
        return {
            name: stateName,
            f: this._invokeLambdaFn(f).bind(this),
        };
    }
    _handleParallel(currentState) {
        this.eventParallelResult = [];
        (0, lodash_1.forEach)(currentState.Branches, (branch) => {
            this.parallelBranch = branch;
            return this.eventForParallelExecution
                ? this.process(branch.States[branch.StartAt], branch.StartAt, this.eventForParallelExecution)
                : null;
        });
        if (currentState.Next) {
            this.process(this.states[currentState.Next], currentState.Next, this.eventParallelResult);
        }
        delete this.parallelBranch;
        this.eventParallelResult = [];
        return;
    }
    _handleChoice(currentState) {
        //push all choices. but need to store information like
        // 1) on which variable need to look: ${variable}
        // 2) find operator: ${condition}
        // 3) find function which will check data: ${checkFunction}
        // 4) value which we will use in order to compare data: ${compareWithValue}
        // 5) find target function - will be used if condition true: ${f}
        const choiceConditional = {
            choice: [],
        };
        (0, lodash_1.forEach)(currentState.Choices, (choice) => {
            var _a;
            const variable = (_a = choice.Variable) === null || _a === void 0 ? void 0 : _a.split('$.')[1];
            const condition = (0, lodash_1.pick)(choice, enum_1.default.supportedComparisonOperator);
            if (!condition) {
                this.cliLog(`Sorry! At this moment we don't support operator '${choice}'`);
                process.exit(1);
            }
            const operator = Object.keys(condition)[0];
            const checkFunction = enum_1.default.convertOperator[operator];
            const compareWithValue = condition[operator];
            if (variable) {
                const choiceObj = {
                    variable,
                    condition,
                    checkFunction,
                    compareWithValue,
                    choiceFunction: choice.Next,
                };
                choiceConditional.choice.push(choiceObj);
            }
        });
        // if exists default function - store it
        if (currentState.Default) {
            choiceConditional.defaultFunction = currentState.Default;
        }
        return choiceConditional;
    }
    _handleWait(currentState, stateName) {
        // Wait State
        // works with parameter: seconds, timestamp, timestampPath, secondsPath;
        return {
            waitState: true,
            f: (event) => __awaiter(this, void 0, void 0, function* () {
                const waitTimer = this._waitState(event, currentState, stateName);
                this.cliLog(`Wait function ${stateName} - please wait ${waitTimer} seconds`);
                yield delay(waitTimer);
                return (e, context, done) => done(null, e || event);
            }),
        };
    }
    _handlePass(currentState) {
        return {
            f: (event) => {
                return (arg1, arg2, cb) => {
                    this.cliLog('!!! Pass State !!!');
                    const eventResult = this._passStateFields(currentState, event);
                    cb(null, eventResult);
                };
            },
        };
    }
    _states(currentState, currentStateName) {
        if ((0, types_1.isType)('Map')(currentState)) {
            return this._handleMap(currentState, currentStateName);
        }
        if ((0, types_1.isType)('Task')(currentState)) {
            return this._handleTask(currentState, currentStateName);
        }
        if ((0, types_1.isType)('Parallel')(currentState)) {
            return this._handleParallel(currentState); // look through branches and push all of them
        }
        if ((0, types_1.isType)('Choice')(currentState)) {
            return this._handleChoice(currentState);
        }
        if ((0, types_1.isType)('Wait')(currentState)) {
            return this._handleWait(currentState, currentStateName);
        }
        if ((0, types_1.isType)('Pass')(currentState)) {
            return this._handlePass(currentState);
        }
        if ((0, types_1.isType)('Fail')(currentState)) {
            const obj = {};
            if (currentState.Cause)
                obj.Cause = currentState.Cause;
            if (currentState.Error)
                obj.Error = currentState.Error;
            this.cliLog('Fail');
            if (!(0, lodash_1.isEmpty)(obj)) {
                this.cliLog(JSON.stringify(obj));
            }
            return Promise.resolve('Fail');
        }
        if ((0, types_1.isType)('Succeed')(currentState)) {
            this.cliLog('Succeed');
            return Promise.resolve('Succeed');
        }
    }
    _passStateFields(currentState, event) {
        if (!currentState.ResultPath) {
            return currentState.Result || event;
        }
        else {
            const variableName = currentState.ResultPath.split('$.')[1];
            if (!currentState.Result) {
                event[variableName] = event;
                return event;
            }
            event[variableName] = currentState.Result;
            return event;
        }
    }
    _runChoice(data, result) {
        let existsAnyMatches = false;
        if (!(data === null || data === void 0 ? void 0 : data.choice))
            return;
        //look through choice and find appropriate
        (0, lodash_1.forEach)(data.choice, (choice) => {
            //check if result from previous function has of value which described in Choice
            const functionResultValue = (0, lodash_1.get)(result, choice.variable);
            if (!(0, lodash_1.isNil)(functionResultValue)) {
                //check condition
                const isConditionTrue = choice.checkFunction(functionResultValue, choice.compareWithValue);
                if (isConditionTrue && choice.choiceFunction) {
                    existsAnyMatches = true;
                    return this.process(this.states[choice.choiceFunction], choice.choiceFunction, result);
                }
            }
        });
        if (!existsAnyMatches && data.defaultFunction) {
            const fName = data.defaultFunction;
            return this.process(this.states[fName], fName, result);
        }
    }
    _waitState(event, currentState, currentStateName) {
        let waitTimer = 0, targetTime, timeDiff;
        const currentTime = dayjs();
        const waitListKeys = ['Seconds', 'Timestamp', 'TimestampPath', 'SecondsPath'];
        const waitField = (0, lodash_1.omit)(currentState, 'Type', 'Next', 'Result');
        const waitKey = Object.keys(waitField)[0];
        if (!waitListKeys.includes(waitKey)) {
            const error = `Plugin does not support wait operator "${waitKey}"`;
            throw new Error(error);
        }
        switch (Object.keys(waitField)[0]) {
            case 'Seconds':
                waitTimer = waitField['Seconds'];
                break;
            case 'Timestamp':
                targetTime = dayjs(waitField['Timestamp']);
                timeDiff = targetTime.diff(currentTime, 'seconds');
                if (timeDiff > 0)
                    waitTimer = timeDiff;
                break;
            case 'TimestampPath':
                const timestampPath = waitField['TimestampPath'].split('$.')[1];
                if (event[timestampPath] === undefined) {
                    const error = `An error occurred while executing the state ${currentStateName}.
                     The TimestampPath parameter does not reference an input value: ${waitField['TimestampPath']}`;
                    throw new Error(error);
                }
                targetTime = dayjs(event[timestampPath]);
                timeDiff = targetTime.diff(currentTime, 'seconds');
                if (timeDiff > 0)
                    waitTimer = timeDiff;
                break;
            case 'SecondsPath':
                const secondsPath = waitField['SecondsPath'].split('$.')[1];
                const waitSeconds = event[secondsPath];
                if (waitSeconds === undefined) {
                    const error = `
                    An error occurred while executing the state ${currentStateName}.
                    The SecondsPath parameter does not reference an input value: ${waitField['SecondsPath']}`;
                    throw new Error(error);
                }
                waitTimer = waitSeconds;
                break;
        }
        return waitTimer;
    }
    addContextObject(states, name, event) {
        if (this.contexts[name])
            return;
        this.contexts[name] = this.createContextObject(states, name, event);
    }
    createContextObject(states, name, originalEvent) {
        let attempt = 0;
        delete originalEvent['default'];
        const cb = (err, result) => {
            var _a, _b, _c, _d, _e, _f, _g;
            if (!(0, types_1.notEmpty)(this.currentState))
                return;
            if (err && !(0, types_1.isType)('Task')(this.currentState)) {
                throw `Error in function "${this.currentStateName}": ${JSON.stringify(err)}`;
            }
            if (err && (0, types_1.isType)('Task')(this.currentState)) {
                const matchingError = ((_a = this.currentState.Retry) !== null && _a !== void 0 ? _a : []).find((condition) => condition.ErrorEquals.includes('HandledError'));
                if (!matchingError)
                    throw `Error in function "${this.currentStateName}": ${JSON.stringify(err)}`;
                attempt += 1;
                if (attempt < ((_b = matchingError.MaxAttempts) !== null && _b !== void 0 ? _b : 0)) {
                    this.addContextObject(states, name, originalEvent);
                    if (matchingError.IntervalSeconds !== undefined && matchingError.IntervalSeconds !== 0) {
                        const backoffRate = (_c = matchingError === null || matchingError === void 0 ? void 0 : matchingError.BackoffRate) !== null && _c !== void 0 ? _c : 2;
                        const fullDelay = attempt === 1
                            ? matchingError.IntervalSeconds
                            : matchingError.IntervalSeconds * (attempt - 1) * backoffRate;
                        console.log(`Delaying ${fullDelay} seconds for execution #${attempt + 1} of state ${name}`);
                        return delay(fullDelay).then(() => this.process(states[name], name, originalEvent));
                    }
                    return this.process(states[name], name, originalEvent);
                }
                const newErr = `Error in function "${this.currentStateName}" after ${attempt} attempts: ${JSON.stringify(this.currentState)} - ${JSON.stringify(err)}`;
                attempt = 0;
                throw newErr;
            }
            attempt = 0;
            this.executionLog(`~~~~~~~~~~~~~~~~~~~~~~~~~~~ ${this.currentStateName} finished ~~~~~~~~~~~~~~~~~~~~~~~~~~~`);
            let state = states;
            if (!(0, types_1.isNotCompletedState)(this.currentState))
                return;
            if (this.parallelBranch && this.parallelBranch.States) {
                state = this.parallelBranch.States;
                if (!((_d = this.currentState) === null || _d === void 0 ? void 0 : _d.Next))
                    this.eventParallelResult.push((_e = serialize(result)) !== null && _e !== void 0 ? _e : {}); // it means the end of execution of branch
            }
            if (this.mapResults && !((_f = this.currentState) === null || _f === void 0 ? void 0 : _f.Next)) {
                this.mapResults.push(serialize(result));
            }
            let nextEvent = result;
            if ((0, types_1.isType)('Task')(this.currentState) && this.currentState.ResultPath) {
                (0, lodash_1.set)(originalEvent, this.currentState.ResultPath.replace(/\$\./, ''), serialize(result));
                nextEvent = originalEvent;
            }
            if ((_g = this.currentState) === null || _g === void 0 ? void 0 : _g.Next) {
                if (this.currentState)
                    this.addContextObject(states, this.currentState.Next, nextEvent !== null && nextEvent !== void 0 ? nextEvent : originalEvent);
                return this.process(state[this.currentState.Next], this.currentState.Next, nextEvent !== null && nextEvent !== void 0 ? nextEvent : {});
            }
        };
        return {
            name,
            attempt,
            cb,
            done: cb,
            succeed: (result) => cb(null, result),
            fail: (err) => cb(err),
        };
    }
    executionLog(log) {
        if (this.detailedLog)
            console.log(log);
    }
}
exports.StepFunctionsOfflinePlugin = StepFunctionsOfflinePlugin;
module.exports = StepFunctionsOfflinePlugin;
//# sourceMappingURL=index.js.map