"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.withWorkers = void 0;
function* arrayGenerator(array) {
    for (let index = 0; index < array.length; index++) {
        const currentValue = array[index];
        yield [currentValue, index, array];
    }
}
function mapItem(mapFn, currentValue, index, array, propagateError = true) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            return {
                status: 'fulfilled',
                value: yield mapFn(currentValue, index, array),
            };
        }
        catch (error) {
            if (propagateError) {
                throw error;
            }
            return {
                status: 'rejected',
                reason: error,
            };
        }
    });
}
function worker(id, gen, mapFn, result, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const [currentValue, index, array] of gen) {
            if (logger) {
                logger(`Worker ${id} --- index ${index} start`);
            }
            result[index] = yield mapItem(mapFn, currentValue, index, array);
            if (logger) {
                logger(`Worker ${id} --- index ${index} ends`);
            }
        }
    });
}
function withWorkers(arr, mapFn, limit = arr.length, logger) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = [];
        if (arr.length === 0) {
            return result;
        }
        const gen = arrayGenerator(arr);
        limit = Math.min(limit, arr.length);
        const workers = new Array(limit);
        for (let i = 0; i < limit; i++) {
            workers.push(worker(i, gen, mapFn, result, logger));
        }
        yield Promise.all(workers);
        return result;
    });
}
exports.withWorkers = withWorkers;
//# sourceMappingURL=worker.js.map